/* ====================================================================
 * Trove - Copyright (c) 1997-2000 Walt Disney Internet Group
 * ====================================================================
 * The Tea Software License, Version 1.1
 *
 * Copyright (c) 2000 Walt Disney Internet Group. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Walt Disney Internet Group (http://opensource.go.com/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Tea", "TeaServlet", "Kettle", "Trove" and "BeanDoc" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact opensource@dig.com.
 *
 * 5. Products derived from this software may not be called "Tea",
 *    "TeaServlet", "Kettle" or "Trove", nor may "Tea", "TeaServlet",
 *    "Kettle", "Trove" or "BeanDoc" appear in their name, without prior
 *    written permission of the Walt Disney Internet Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE WALT DISNEY INTERNET GROUP OR ITS
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * For more information about Tea, please see http://opensource.go.com/.
 */

package com.go.trove.util.tq;

import java.util.EventObject;
import java.util.Date;

/******************************************************************************
 * An event that contains information from a {@link TransactionQueue}.
 * TransactionQueueEvents can be received by implementing a
 * {@link TransactionQueueListener}.
 *
 * @author Brian S O'Neill
 * @version
 * <!--$$Revision: 1.1 $-->, <!--$$JustDate:-->  9/25/00 <!-- $-->
 */
public class TransactionQueueEvent extends EventObject {
    private Transaction mTransaction;
    private Throwable mThrowable;
    private long mTimestampMillis;
    private long mStageDuration;

    private transient Date mTimestamp;

    public TransactionQueueEvent(TransactionQueue source, 
                                 Transaction transaction) {
        super(source);
        mTransaction = transaction;
        mTimestampMillis = System.currentTimeMillis();
    }

    public TransactionQueueEvent(TransactionQueue source, 
                                 Transaction transaction,
                                 Throwable throwable) {
        super(source);
        mTransaction = transaction;
        mThrowable = throwable;
        mTimestampMillis = System.currentTimeMillis();
    }

    /**
     * Copies the given event, but gives it a new timestamp and computes a
     * stage duration.
     */
    public TransactionQueueEvent(TransactionQueueEvent event) {
        super(event.getSource());
        mTransaction = event.mTransaction;
        mThrowable = event.mThrowable;
        mStageDuration = (mTimestampMillis = System.currentTimeMillis()) -
            event.mTimestampMillis;
    }

    /**
     * Copies the given event, but gives it a different throwable, a new
     * timestamp and computes a stage duration.
     */
    public TransactionQueueEvent(TransactionQueueEvent event,
                                 Throwable throwable) {
        super(event.getSource());
        mTransaction = event.mTransaction;
        mThrowable = throwable;
        mStageDuration = (mTimestampMillis = System.currentTimeMillis()) -
            event.mTimestampMillis;
    }

    public TransactionQueue getTransactionQueue() {
        return (TransactionQueue)getSource();
    }

    /**
     * Returns the date and time of this event.
     */
    public Date getTimestamp() {
        if (mTimestamp == null) {
            mTimestamp = new Date(mTimestampMillis);
        }
        return mTimestamp;
    }

    /**
     * Returns the date and time of this event as milliseconds since 1970.
     */
    public long getTimestampMillis() {
        return mTimestampMillis;
    }

    /**
     * Returns the amount of time, in milliseconds, that the transaction was
     * in a TransactionQueue stage. The implied meaning of this value depends
     * on how this event was passed to a listener.
     *
     * @see TransactionQueueListener
     */
    public long getStageDuration() {
        return mStageDuration;
    }

    /**
     * Returns the transaction that was being processed by the
     * TransactionQueue.
     */
    public Transaction getTransaction() {
        return mTransaction;
    }

    /**
     * Returns the exception that occurred while servicing the transaction, if
     * applicable.
     */
    public Throwable getThrowable() {
        return mThrowable;
    }
}
